/**
This applet displays mobile coloured polygons, or bots.
Each bot is controlled by a separate thread (of class Bot).
Each bot senses where it is, where the mouse is,
where the other bots are, and where the container walls are.

Parameters:
	numBots (the number of bots to be displayed; default = 1)

Globals:
	offscreenImage, offscreenGraphics,
	bots[], numBots,
	screenUpdate, pauseScreen

Public Methods:
	init()
	start()
	stop()
	run()
	mouseDown()
	mouseMove()

Flow of control:
	The Soup applet sets up an offscreen bitmap, creates the bots, then
creates a screenUpdate thread to keep refreshing the screen. In the thread it
continually polls the bots to find out where they are (having previously
determined their shape and colour and so on) and redraws them in their new
positions. Each bot runs autonomously in its own thread, first choosing its
shape, colour and so on, then waiting until all bots are alive before
polling them to get handles on their threads. It then loops continuously,
looking at where the mouse is (data supplied to it by the Soup applet) and
runs away from the mouse if it's moving. Every now and then it looks at the
other bots to see where they are and if they're moving closer. If so it
runs away from them.
*/

import java.awt.Image;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Event;
import java.awt.Point;
import java.awt.Polygon;
import java.util.Random;
import java.applet.Applet;

//////////////////////////////////////////////////////////////////////
public class Soup extends Applet implements Runnable
	{
	Image offscreenImage;        //offscreen bitmap
	Graphics offscreenGraphics;  //its graphics context

	Bot bots[];                  //array of bot threads
	int numBots = 1;             //default number of bots to be created

	Thread screenUpdate = null;  //screen update thread
	boolean pauseScreen = false; //has my user asked me to pause?

	///////////////////////////////////////////////////////////////
	public void init()
		{
		/**
		Set up the screen size, the offscreen bitmap for double-buffering,
		the random number stream, and parse the number of bots requested,
		then set up and start all the bots.

		Globals: offscreenImage, offscreenGraphics, bots[], numBots
		*/

		//make sure my display is at least 200X200 pixels
		//[Note: size() is defined on all Components (of which Applet is one)]
		int screenWidth = 200;
		int screenHeight = 200;
		if (this.size().width  >= screenWidth)  screenWidth = this.size().width;
		if (this.size().height >= screenHeight) screenHeight = this.size().height;
		resize(screenWidth, screenHeight);

		//create an offscreen bitmap and its graphics context
		//and flush the offscreen image to white
		//[Note: createImage() is defined on all Components]
		offscreenImage = this.createImage(screenWidth, screenHeight);
		offscreenGraphics = offscreenImage.getGraphics();
		offscreenGraphics.setColor(Color.white);
		offscreenGraphics.fillRect(0, 0, screenWidth, screenHeight);

		//get a new pseudo-random number generator;
		//it's created here because if it's created in the child threads
		//it will produce the same stream for all threads created at the
		//same time
		Random PRNG = new Random();

		//get the number of bots to create from the html file;
		//use the default if the number is absent, zero or less, or malformed
		String s = getParameter("numBots");
		int n = 0;
		if (s != null)
			{
			try {n = Integer.parseInt(s);}
			catch (NumberFormatException ignored)
				{
				System.err.println("Couldn't parse parameter " +
									"`numBots'; using default instead.");
				}
			}
		if (n > 0) numBots = n;

		//create numBots bots;
		//pass each one a pointer to me and the common PRNG
		bots = new Bot[numBots];
		for (int i = 0; i < numBots; i++)
			{
			bots[i] = new Bot(this, PRNG);
			bots[i].setName("Bot" + i);
			bots[i].start();
			Thread.currentThread().yield();
			}

		//let the bots know that all of them are alive now
		//so that they can each find each other
		for (int i = 0; i < numBots; i++)
			{
			bots[i].startUp();
			Thread.currentThread().yield();
			}
		}

	///////////////////////////////////////////////////////////////
	public void start()
		{
		pauseScreen = false;

		//create a thread simply to do periodic screen updates
		if (screenUpdate == null)
			{
			screenUpdate = new Thread(this);
			screenUpdate.start();

			//resume the child threads
			for (int i = 0; i < numBots; i++)
				if (bots[i].isAlive())
					bots[i].resume();
			}
		}

	///////////////////////////////////////////////////////////////
	public void stop()
		{
		if ((screenUpdate != null) && (screenUpdate.isAlive()))
			{
			//pause the child threads
			for (int i = 0; i < numBots; i++)
				if (bots[i].isAlive())
					bots[i].suspend();

			//and stop the screen refresh thread
			screenUpdate.stop();
			//zero it out so that the garbage collector can reclaim it
			screenUpdate = null;
			}

		//my user may also have simply closed my window, so set pauseScreen
		pauseScreen = true;
		}

	///////////////////////////////////////////////////////////////
	public void run()
		{
		/**
		Refresh my screen roughly 25 times a second.
		*/

		Color[] botColor = new Color[numBots];
		Polygon[] botShape = new Polygon[numBots];
		Polygon[] oldShape = new Polygon[numBots]; //last known botpositions
		Polygon newShape; //temporarily holds newest estimated position

		//first get each bot's colour and shape
		for (int i = 0; i < numBots; i++)
			{
			botColor[i] = bots[i].mycolour();
			botShape[i] = bots[i].myshape();
			oldShape[i] = bots[i].myshape();
			}

		Point botPosition; //just for shorthand in the following loop
		while (!pauseScreen)
			{
			for (int i = 0; i < numBots; i++)
				{
				//erase bot i's last painted position
				offscreenGraphics.setColor(Color.white);
				offscreenGraphics.fillPolygon(oldShape[i]);

				//get bot i's newest position
				botPosition = bots[i].botpos();
				newShape = new Polygon();
				for (int j = 0; j < botShape[i].npoints; j++)
					{
					newShape.addPoint(botShape[i].xpoints[j] + botPosition.x,
									botShape[i].ypoints[j] + botPosition.y);
					}

				//draw bot i's newest position
				offscreenGraphics.setColor(botColor[i]);
				offscreenGraphics.fillPolygon(newShape);

				//remember bot i's last painted position
				oldShape[i] = newShape;
				}

			repaint();
			//instead of calling repaint() here,
			//could get the applet's graphics context directly
			//and bitblit the offscreen bitmap directly to the screen
			//[Note: getGraphics() is defined on all Components]
			//Graphics g = this.getGraphics();
			//if (g != null)	//just in case this applet isn't on screen now
				//{
				//g.drawImage(offscreenImage, 0, 0, this);
				//g.dispose();
				//}

			try { Thread.sleep(40); }
			catch (InterruptedException ignored) {}
			}
		}

	///////////////////////////////////////////////////////////////
	public void update(Graphics g)
		{
		paint(g);
		}

	///////////////////////////////////////////////////////////////
	public void paint(Graphics g)
		{
		g.drawImage(offscreenImage, 0, 0, this);
		}

	///////////////////////////////////////////////////////////////
	public boolean mouseDown(Event event, int mouseX, int mouseY)
		{
		/**
		My user has clicked the mouse in my window, so toggle pauseScreen
		*/

		pauseScreen = !pauseScreen;

		if (!pauseScreen)
			{
			//create a thread simply to do periodic screen updates
			if (screenUpdate == null)
				{
				screenUpdate = new Thread(this);
				screenUpdate.start();

				//resume the child threads
				for (int i = 0; i < numBots; i++)
					if (bots[i].isAlive())
						bots[i].resume();
				}
			}

		return true;
		}

	///////////////////////////////////////////////////////////////
	public boolean mouseMove(Event event, int mouseX, int mouseY)
		{
		/**
		Whenever the mouse moves tell the bots the current mouse position

		Calls: Bot.getMouseInfo()

		Globals: numBots, bots[]
		*/

		//instead of this stuff, how about putting the mouse moves in
		//an instance variable and having the bots access that array
		//whenever they want to? Would need to make access to it synchronized.

		Point mousePos = new Point(mouseX, mouseY);
		for (int i = 0; i < numBots; i++)
			//update code is synchronized on each bot
			//to prevent changing the mouse's position
			//while that particular bot happens to be reading it
			synchronized (bots[i])
				{
				bots[i].getMouseInfo(mousePos);
				}

		return true;
		}

	}//////////end of class Soup///////////
